"""Don't import this module directly."""

from typing import Optional, overload

from .. import utils
from .mod import Mod
from .. import spec


class Repository:
    """A repository."""

    address: Optional[str]
    """The address of the repository."""
    data: spec.RRepository
    """The raw data of the repository."""

    @overload
    def __init__(self, address: Optional[str], data: spec.RRepository):
        """Initializes a repository with an address and data."""

    @overload
    def __init__(self, address: str):
        """Initializes a repository with an address. The data will be fetched."""

    @overload
    def __init__(self, address: Optional[str], data: dict):
        """Initializes a repository with an address and data.
        The data will be converted to a spec.RRepository."""

    def __init__(self, address, data=None):
        if isinstance(data, dict):
            data = spec.RRepository.from_dict(data)
        self.address = address
        self.data = data
        if data is None:
            self.update()
        if not isinstance(self.data, spec.RRepository):
            raise ValueError("Invalid data type")
        if self.data.spec_version != 1:
            raise ValueError("Unsupported spec version", self.data.spec_version)

    def get_mod(self, id_: str) -> Optional[Mod]:
        """Gets a mod by its ID."""
        for mod in self.data.mods:
            if mod.id == id_:
                return Mod(mod)
        return None

    def has_mod(self, id_: str) -> bool:
        """Checks if a mod exists in the repository."""
        return self.get_mod(id_) is not None

    @property
    def root_id(self) -> str:
        """The root ID of the repository."""
        return self.data.root_id

    def update(self):
        """Updates the repository data."""
        self.data = utils.fetch_repository(self.address)



"""Don't import this module directly."""

from .types import Repository
from .utils import get_request

AUTOREPO_URL = "https://crm-repo.jojojux.de"


def get_all_repos():
    """Fetches all known repositories from the autorepo server."""
    address = f"{AUTOREPO_URL}/repo_mapping.json"
    response = get_request(address)
    return [Repository(repo) for repo in response["repos"].values()]



"""Don't import this module directly."""

from typing import Optional, overload

from .. import spec
from .mod import Mod
from .repository import Repository


class RepositoryPool:
    """A pool of repositories."""

    repositories: dict[str, Repository]
    """The repositories in the pool."""

    def __init__(self):
        self.repositories = {}

    @overload
    def add_repository(self, address: str):
        """Adds a repository to the pool by providing the address."""

    @overload
    def add_repository(self, repo: spec.RRepository):
        """Adds a repository to the pool that has already been fetched."""

    @overload
    def add_repository(self, repo: Repository):
        """Adds a repository to the pool."""

    @overload
    def add_repository(self, data: dict[str, str]):
        """Adds a repository to the pool."""

    def add_repository(self, repo):
        """Above"""
        if isinstance(repo, str):
            repo = Repository(repo)
        if isinstance(repo, dict):
            repo = spec.RRepository.from_dict(repo)
        if isinstance(repo, spec.RRepository):
            repo = Repository(None, repo)
        self.repositories[repo.root_id] = repo

    def get_repository(self, root_id: str) -> Repository:
        """Gets a repository by its root ID."""
        return self.repositories[root_id]

    def get_mod(self, id_: str) -> Optional[Mod]:
        """Gets a mod by its ID."""
        for repo in self.repositories.values():
            mod = repo.get_mod(id_)
            if mod is not None:
                return mod
        return None

    def has_mod(self, id_: str) -> bool:
        """Checks if a mod exists in the pool."""
        return self.get_mod(id_) is not None

    def where_is(self, id_: str) -> Optional[Repository]:
        """Gets the repository that contains the mod."""
        for repo in self.repositories.values():
            if repo.has_mod(id_):
                return repo
        return None

    def update_repositories(self):
        """Updates all repositories in the pool."""
        for repo in self.repositories.values():
            repo.update()

    def __getitem__(self, key: str) -> Repository:
        return self.get_repository(key)


@overload
def make_pool(repos: list[Repository]) -> RepositoryPool:
    """Creates a repository pool from a list of repositories."""


@overload
def make_pool(*repos: Repository) -> RepositoryPool:
    """Creates a repository pool from multiple repositories."""


def make_pool(*repos):
    """Above"""
    if len(repos) == 1 and isinstance(repos[0], list):
        repos = repos[0]
    pool = RepositoryPool()
    for repo in repos:
        if not isinstance(repo, Repository):
            raise TypeError("Invalid repository", repo)
        pool.add_repository(repo)
    return pool



"""Don't import this module directly."""

from .. import spec
from .dependency import Dependency


class Mod:
    """This class represents a mod.

    Raises:
        ValueError: If the data is not of type spec.RMod
    """

    meta: spec.RMod
    """The raw data of the mod."""

    def __init__(self, data: spec.RMod):
        self.meta = data
        if not isinstance(self.meta, spec.RMod):
            raise ValueError("Invalid data type")

    @property
    def original_ext(self) -> dict:
        """This holds the original mod.ext data."""
        return self.meta.ext.to_dict()

    @property
    def id(self) -> str:
        """The ID of the mod."""
        return self.meta.id

    @property
    def depends(self) -> list[Dependency]:
        """The dependencies of the mod."""
        return [Dependency(dep) for dep in self.meta.deps]

    @property
    def suggests(self) -> list[Dependency]:
        """The suggestions of the mod."""
        if self.meta.ext.suggests is not None:
            return [Dependency(dep) for dep in self.meta.ext.suggests]



"""Don't import this module directly."""

from dataclasses import dataclass

from dataclasses_json import LetterCase

from .common_modext import CommonModExt
from .dependency import RDependency

from dataclasses_hjson import DataClassHjsonMixin


@dataclass
class RMod(DataClassHjsonMixin):
    """Raw mod data. This is used for deserialization."""

    id: str
    """Mod ID. This is in the format of group.id, like `com.example.mod`."""
    name: str
    """Mod name."""
    desc: str
    """A short description of the mod."""
    authors: list[str]
    """A list of the authors' names of the mod."""
    version: str
    """The version of the mod."""
    game_version: str
    """The version of the game that the mod is compatible with."""
    url: str
    """The download URL of the mod's jar."""
    deps: list[RDependency]
    """A list of dependencies of the mod."""
    ext: CommonModExt
    """Some optional additional info of the mod."""



"""Don't import this module directly."""

from functools import lru_cache

import hjson
import requests

from . import spec


@lru_cache(maxsize=128)
def get_request(address: str) -> dict:
    """Performs a GET request to the given address and returns the HJSON response."""
    response = requests.get(address, timeout=5)
    return hjson.loads(response.text, object_pairs_hook=dict)


def fetch_repository(address: str) -> spec.RRepository:
    """Fetches a repository from the given address."""
    data = get_request(address)
    return spec.RRepository.from_dict(data)



"""A module for handling versions."""

from dataclasses import dataclass
from enum import Enum
from typing import Optional


class VersionEndMode(Enum):
    """The mode of the end of a version range. This can be inclusive or exclusive."""

    INCLUSIVE = "inclusive"
    EXCLUSIVE = "exclusive"


DONTCARE = VersionEndMode.EXCLUSIVE
"""This is used to represent the end of a version range which's version is not specified."""


@dataclass
class Version:
    """A version."""

    major: int
    """The major version."""
    minor: int
    """The minor version."""
    patch: int
    """The patch version."""

    @staticmethod
    def from_string(version: str) -> "Version":
        """Create a Version object from a string.
        The string should be in the format of `major.minor.patch`."""
        parts = version.split(".")
        if len(parts) < 2:
            raise ValueError("Invalid version string")
        if len(parts) > 3:
            raise ValueError("Invalid version string")
        major = int(parts[0])
        minor = int(parts[1])
        patch = int(parts[2]) if len(parts) == 3 else 0
        return Version(major, minor, patch)

    def to_string(self) -> str:
        """Convert the version to a string in the format of `major.minor.patch`."""
        return f"{self.major}.{self.minor}" + (
            f".{self.patch}" if self.patch is not None else ""
        )

    def __eq__(self, other: "Version") -> bool:
        return (
            self.major == other.major
            and self.minor == other.minor
            and self.patch == other.patch
        )

    def __lt__(self, other: "Version") -> bool:
        if self.major < other.major:
            return True
        if self.major > other.major:
            return False
        if self.minor < other.minor:
            return True
        if self.minor > other.minor:
            return False
        if self.patch < other.patch:
            return True
        return False

    def __le__(self, other: "Version") -> bool:
        return self == other or self < other

    def __gt__(self, other: "Version") -> bool:
        return not self <= other

    def __ge__(self, other: "Version") -> bool:
        return not self < other

    def __ne__(self, other: "Version") -> bool:
        return not self == other

    def __hash__(self) -> int:
        return hash((self.major, self.minor, self.patch))

    def __repr__(self) -> str:
        return f"Version({self.major}, {self.minor}, {self.patch})"

    def __str__(self) -> str:
        return self.to_string()


@dataclass
class VersionRange:
    """A range of versions."""

    lower: Optional[Version]
    """The lower bound of the range."""
    lower_mode: VersionEndMode
    """The mode of the lower bound of the range."""
    upper: Optional[Version]
    """The upper bound of the range."""
    upper_mode: VersionEndMode
    """The mode of the upper bound of the range."""

    def __post_init__(self):
        if (
            self.lower is not None
            and self.upper is not None
            and self.lower > self.upper
        ):
            raise ValueError("Invalid range")

    @staticmethod
    def from_string(range_: str) -> "VersionRange":
        """Create a VersionRange object from a string.
        The string should be in the format of `[lower,upper]`. Examples: `[1.0,2.0.1)`, `(1.0,]`.
        """
        if range_.startswith("["):
            lower_mode = VersionEndMode.INCLUSIVE
        elif range_.startswith("("):
            lower_mode = VersionEndMode.EXCLUSIVE
        else:
            return VersionRange(Version.from_string(range_), DONTCARE, Version.from_string(range_), DONTCARE)
        if range_.endswith("]"):
            upper_mode = VersionEndMode.INCLUSIVE
        elif range_.endswith(")"):
            upper_mode = VersionEndMode.EXCLUSIVE
        else:
            raise ValueError("Invalid range string")
        range_ = range_[1:-1]
        parts = range_.split(",")
        if len(parts) != 2:
            raise ValueError("Invalid range string")
        lower = Version.from_string(parts[0]) if parts[0] else None
        upper = Version.from_string(parts[1]) if parts[1] else None
        return VersionRange(lower, lower_mode, upper, upper_mode)

    def to_string(self) -> str:
        """Convert the version range to a string in the format of `[lower,upper]`.
        Examples: `[1.0,2.0.1)`, `(1.0,]`."""
        if (
            self.lower is not None
            and self.upper is not None
            and self.lower == self.upper
        ):
            return self.lower.to_string()
        return (
            (
                ("[" if self.lower_mode == VersionEndMode.INCLUSIVE else "(")
                if self.lower is not None
                else "("
            )
            + (self.lower.to_string() if self.lower else "")
            + ","
            + (self.upper.to_string() if self.upper else "")
            + (
                ("]" if self.upper_mode == VersionEndMode.INCLUSIVE else ")")
                if self.upper is not None
                else ")"
            )
        )

    def contains(self, version: Version) -> bool:
        """Check if the version is in the range."""
        if self.lower is not None:
            if self.lower_mode == VersionEndMode.INCLUSIVE:
                if version < self.lower:
                    return False
            else:
                if version <= self.lower:
                    return False
        if self.upper is not None:
            if self.upper_mode == VersionEndMode.INCLUSIVE:
                if version > self.upper:
                    return False
            else:
                if version >= self.upper:
                    return False
        return True

    def __contains__(self, version: Version) -> bool:
        return self.contains(version)

    def __str__(self) -> str:
        return self.to_string()


def range_from_maven_string(version: str):
    """Create a VersionRange object from a Maven version string."""
    if version == "*":
        return VersionRange(None, DONTCARE, None, DONTCARE)
    if version.startswith(">="):
        lower_mode = VersionEndMode.INCLUSIVE
        lower = Version.from_string(version[2:])
        upper = None
        upper_mode = DONTCARE
    elif version.startswith(">"):
        lower_mode = VersionEndMode.EXCLUSIVE
        lower = Version.from_string(version[1:])
        upper = None
        upper_mode = DONTCARE
    elif version.startswith("<="):
        lower_mode = DONTCARE
        lower = None
        upper = Version.from_string(version[2:])
        upper_mode = VersionEndMode.INCLUSIVE
    elif version.startswith("<"):
        lower_mode = DONTCARE
        lower = None
        upper = Version.from_string(version[1:])
        upper_mode = VersionEndMode.EXCLUSIVE
    else:
        lower_mode = VersionEndMode.INCLUSIVE
        lower = Version.from_string(version)
        upper = Version.from_string(version)
        upper_mode = VersionEndMode.INCLUSIVE
    return VersionRange(lower, lower_mode, upper, upper_mode)
